## Backtrace-Source-Code

- Backtrace service for reading source code directly from version control

## Development
- `yarn watch-build` in one terminal
- `NODE_ENV=development yarn watch-start` in a second terminal
- Alternatively: `NODE_ENV=development yarn watch-debug` in the second terminal to enable debugging via chrome devtools

## On-prem & hosted coroner
- Due to the potentially sensitive nature of source code, some customers with hosted coroner may want to run backtrace-source-code on-prem
- This goes against the model of most of our other services, because coroner generally needs to be able to communicate with a service in order to fully handshake with it
  - Without a full handshake, the service will not appear in the services section of the config object returned by coroner at /api/config
  - Are there other consequences of not running a full handshake?
- In order for coroner to communicate with backtrace-source-code, it must have a network route to backtrace-source-code running on-prem at port 6451
- However, this requires the customer to open up their firewall to whitelist our hosted coroner's IP address to enable it to communicate with backtrace-source-code running on-prem at 6451
- This may not be possible for certain high-security clients
- Consequently, backtrace-source-code supports a standalone mode which does not authenticate with coroner

## Standalone mode
- __Standalone mode should only be run in high security network environments!__
  - Only coroner knows who the valid users are and what their role is within a project (admin or member)
  - Only coroner can verify that a coroner token is currently valid and belongs to a signed-in user
- Because authentication with coroner is disabled, backtrace-source-code will respond to all requests for source code lookups without verifying that the requester has access. __In standalone mode, security over source code is solely provided at the network / firewall level.__
- Standalone mode can be enabled by setting NODE_ENV environment variable to 'standalone'
- In this mode, it does not require super-user privileges to run
  - The database is stored in a subfolder named "database" under the backtrace-source-code project root
  - The sourcecode.conf file (if present) is read from the backtrace-source-code project root
  - The service does not handshake with coroner, since the availability of a network path from coroner to backtrace-source-code may not be available through the firewall
- Window variable required
  - Because there is no handshake with coroner, coroner will not advertise the service to the web console in the services section of /api/config
  - As a result, the web console needs to know where to fetch source code from
  - We use _BACKTRACE_SERVICE_OVERRIDES window variable for this
  - The format is an array of two-element arrays: [["_PROJECT_NAME_", {name: "_SERVICE_NAME", endpoint: "_ENDPOINT_URL"}]]
  - Quotes must be escaped in order to make this work as a window variable. Use the following as a template:
    - "_BACKTRACE_SERICE_OVERRIDES":  "[[\"*\",{\"name\":\"sourcecode\",\"endpoint\":\"https://10.0.0.3:4312\"}]]"
  - The above sets all projects (using the "*" wildcard) to use https://10.0.0.3:4312 as the sourcecode service URL
  - Typically, it's enough to edit the above string and it will work, but more complex setups are supported
    - We can have more than one entry
    - We can set a custom endpoint on a per-project basis
    -  To get a properly escaped string, run this in your browser console and strip off the outer quotes from the output (using google chrome): JSON.stringify(JSON.stringify([["amadden",{"name":"sourcecode","endpoint":"https://localhost:4312"}]]))

## Running backtrace-source-code service on-prem
- Installing as a Linux package
  - .deb and .rpm files available from Backtrace's packagecloud repository at https://packagecloud.io/backtrace/packages
- Installing as a nodejs project (Mac OS, Linux, & Windows (untested))
  - Download the latest package file from packagecloud
    - Filename will look like backtrace-source-code-{version}.tgz
  - Install to a local folder
    - `tar xzvf backtrace-source-code-1.2.1.tgz`
    - `cd package`
    - `npm install`
      - If the above step fails while building the nodegit dependency:
        - Redhat / Centos
          - yum install curl openssl-devel libcurl-devel python make glibc-devel gcc gcc-c++ krb5-devel
          - npm rebuild --build-from-source nodegit
        - Ubuntu / Debian
          - Install dependencies listed here: https://github.com/nodegit/nodegit/tree/v0.26.2#getting-started
          - npm rebuild --build-from-source nodegit
    - Starting the service
      - For standalone mode, use `npm run local-start`
      - For production mode, integrated with coronerd, use `npm run start` or `npm start` for short
- Perforce support
  - Support for Perforce repositories requires that the p4 command-line utility be installed and available
    - Without the helix command line client (p4 command-line utility), you will be unable to add a Perforce server in the source code config section of the web console
    - Download link & install instructions available at: https://www.perforce.com/downloads/helix-command-line-client-p4
    - Once it is installed it should be available in the PATH of the user that runs `npm run local-start`, or the full path to the p4 command-line binary can be specified via:
      - environment variable P4CMD
      - within sourcecode.conf, under the P4CMD key-value pair
        - sourcecode.conf is a json formatted file which should be placed within your backtrace-source-code installation's root folder

## Packaging backtrace-source-code for a standalone environment
- Use `npm run pack` to create a backtrace-source-code-{version}.tgz file
- Upload to Backtrace's public packagecloud repository at https://packagecloud.io/backtrace/

## Schema migrations
Run all migrations which haven't been run yet:
- `NODE_ENV=development yarn exec -- knex migrate:latest`
Create a new migration:
- `NODE_ENV=development yarn exec -- knex migrate:make add_position_to_repositories`
Roll back one migration:
- `NODE_ENV=development yarn exec -- knex migrate:down`
More commands:
- `NODE_ENV=development yarn exec -- knex --help`
On a production server:
- `mrashkovsky@source-code01:/opt/backtrace/services/backtrace-source-code$ sudo NODE_ENV=production npm run-script exec -- ./node_modules/.bin/knex migrate:latest`

## Tasks
Task files can be used to migrate data, after doing a schema migration.
For example, after adding a "position" column to the repositories table, we might want to set it to a proper value for every row in the repositories table.
An example of this kind of task is in `tasks/add_position_to_repositories.js`
To run it, your current working directory _must_ be the project root. It will not work otherwise.
Then run:
- `NODE_ENV=development node ./tasks/add_position_to_repositories.js`
On a production server:
- `mrashkovsky@source-code01:/opt/backtrace/services/backtrace-source-code$ sudo NODE_ENV=production npm run-script exec -- node ./tasks/copy_repositories_over.js`

## Perforce
For testing purposes, you may need to convert a git repo to a perforce repo
 - The following steps were adapted from
   - https://git-scm.com/docs/git-p4
   - https://stackoverflow.com/questions/2428137/how-to-rebase-one-git-repository-onto-another-one/2428339#2428339
 - You must have an existing workspace with at least one file in your perforce server's folder
   - cd ~/p4workspace
   - mkdir main
   - echo "test" >> main/tempfile.txt
   - p4 add main/tempfile
   - p4 submit
 - Import the above perforce directory into a git repo
   - cd ~/sourcecode/
   - git p4 clone //depot/main
   - cd main
 - Make note of the first commit of the perforce directory / git repo
   - git rev-list --max-parents=0 HEAD
     - 22e730c4ab0088a9a502bbd7c5eaacb9bf2bdfd1'
 - Add the main git repo as a remote
   - git remote add bsc ~/sourcecode/backtrace-source-code
   - git fetch bsc
 - Make note of the first commit in the backtrace-source-code repo
   - git checkout bsc/master
   - git rev-list --max-parents=0 HEAD
     - 9b206ca22a39849b1ffe377034d99220c330b674
 - Since we're dealing with two different repos, we have to graft the history of one on top of the other
   - we want to stack backtrace-source-code repo's history on top of the perforce repo's history
   - echo '9b206ca22a39849b1ffe377034d99220c330b674 22e730c4ab0088a9a502bbd7c5eaacb9bf2bdfd1' > .git/info/grafts
   - git replace --convert-graft-file
   - check git log to make sure the first commit is now
   - You should still be in bsc/master, let's rename it to master
     - git branch -d master
     - git checkout master
 - Now we submit this new history back to the perforce folder
   - EDITOR='vim -c ":wq"' git p4 submit
     - This command goes through each commit in your branch and creates an equivalent changeset in perforce
     - It'll ask you to edit the commit message each time, so we include that EDITOR environment variable to automatically confirm  and save their default commit message
 - If you have any commits with empty directories, the above command will complain about adding it as an "empty file". Delete that folder from your history with the following steps
   - The following steps are adapted from https://stackoverflow.com/a/32886427/2569381
     - FILTER_BRANCH_SQUELCH_WARNING=1 git filter-branch --index-filter 'git rm -rf --cached --ignore-unmatch THE_FOLDER_TO_REMOVE/' --prune-empty --tag-name-filter cat -- --all
     - git for-each-ref --format="%(refname)" refs/original/ | xargs -n 1 git update-ref -d
     - rm -Rf .git/logs .git/refs/original
     - git gc --prune=all --aggressive
