const fs = require("fs");
const path = require("path");
const { BacktraceService, Logger, PackageUtils } = require("backtrace-service");

const log = Logger.getLogger();
let configDb;
if (process.env.NODE_ENV === "test") {
  configDb = {
    client: "sqlite3",
    connection: {
      filename: `${process.cwd()}/database/test/sourcecode.test.db`,
    },
  };
} else {
  configDb = BacktraceService.readConfiguration(
    "sourcecode",
    6451,
    undefined,
    false
  ).database;
}

// Check db directory if it does not already exist
const directory = path.dirname(configDb.connection.filename);
if (!fs.existsSync(directory)) {
  fs.mkdirSync(directory, { recursive: true }, (err) => {
    if (err) {
      log.error(`Error: Database directory ${directory} cannot be created.`);
      throw err;
    }
  });
}

const defaults = {
  migrations: {
    directory: `${PackageUtils.packageDir}/dist/migrations`,
  },
  useNullAsDefault: true,
};

module.exports = Object.assign(defaults, configDb);
